from typing import TYPE_CHECKING

from . import conf as conf
from . import extras as extras
from ._cli import cli as cli
from ._fields import MISSING as MISSING
from ._instantiators import (
    UnsupportedTypeAnnotationError as UnsupportedTypeAnnotationError,
)

# Deprecated interface.
if not TYPE_CHECKING:
    from ._deprecated import *  # noqa


# TODO: this should be synchronized automatically with the pyproject.toml.
__version__ = "0.7.1"
